
# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'rootB CNOT CP CZ FLOAT FSIM H ISWP M MEASURE NEWLINE QREG RX RXY RY RZ S SD SISWP SSWP SWP T TD X X2M X2P XY XY2M XY2P XYARB Y Y2M Y2P Z Z2M Z2P Z4M Z4Proot : programprogram : instruction NEWLINE program\n        | instruction\n        | NEWLINE program\n        |\n        instruction : MEASURE qlist\n        | M qlist\n        | B qlistqlist : QREG qlist\n        | QREG\n        instruction : XYARB QREG FLOAT FLOAT\n        | RXY QREG FLOAT FLOATinstruction : RX QREG FLOAT\n        | RY QREG FLOATinstruction : XY QREG FLOAT\n        | XY2P QREG FLOAT\n        | XY2M QREG FLOAT\n        | RZ QREG FLOATinstruction : X QREG\n        | X2P QREG\n        | X2M QREG\n        | Y QREG\n        | Y2P QREG\n        | Y2M QREG\n        | Z QREG\n        | Z2P QREG\n        | Z2M QREG\n        | Z4P QREG\n        | Z4M QREG\n        | S QREG\n        | SD QREG\n        | T QREG\n        | TD QREG\n        | H QREGinstruction : CZ QREG QREG\n        | CNOT QREG QREG\n        | SWP QREG QREG\n        | SSWP QREG QREG\n        | ISWP QREG QREG\n        | SISWP QREG QREGinstruction : CP QREG QREG FLOAT\n        | FSIM QREG QREG FLOAT'
    
_lr_action_items = {'NEWLINE':([0,3,4,40,42,43,44,45,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,79,82,83,84,85,86,87,88,89,90,91,92,93,96,97,98,99,],[4,40,4,4,-6,-10,-7,-8,-19,-20,-21,-22,-23,-24,-25,-26,-27,-28,-29,-30,-31,-32,-33,-34,-9,-13,-14,-15,-16,-17,-18,-35,-36,-37,-38,-39,-40,-11,-12,-41,-42,]),'$end':([0,1,2,3,4,40,41,42,43,44,45,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,78,79,82,83,84,85,86,87,88,89,90,91,92,93,96,97,98,99,],[-5,0,-1,-3,-5,-5,-4,-6,-10,-7,-8,-19,-20,-21,-22,-23,-24,-25,-26,-27,-28,-29,-30,-31,-32,-33,-34,-2,-9,-13,-14,-15,-16,-17,-18,-35,-36,-37,-38,-39,-40,-11,-12,-41,-42,]),'MEASURE':([0,4,40,],[5,5,5,]),'M':([0,4,40,],[6,6,6,]),'B':([0,4,40,],[7,7,7,]),'XYARB':([0,4,40,],[8,8,8,]),'RXY':([0,4,40,],[9,9,9,]),'RX':([0,4,40,],[10,10,10,]),'RY':([0,4,40,],[11,11,11,]),'XY':([0,4,40,],[12,12,12,]),'XY2P':([0,4,40,],[13,13,13,]),'XY2M':([0,4,40,],[14,14,14,]),'RZ':([0,4,40,],[15,15,15,]),'X':([0,4,40,],[16,16,16,]),'X2P':([0,4,40,],[17,17,17,]),'X2M':([0,4,40,],[18,18,18,]),'Y':([0,4,40,],[19,19,19,]),'Y2P':([0,4,40,],[20,20,20,]),'Y2M':([0,4,40,],[21,21,21,]),'Z':([0,4,40,],[22,22,22,]),'Z2P':([0,4,40,],[23,23,23,]),'Z2M':([0,4,40,],[24,24,24,]),'Z4P':([0,4,40,],[25,25,25,]),'Z4M':([0,4,40,],[26,26,26,]),'S':([0,4,40,],[27,27,27,]),'SD':([0,4,40,],[28,28,28,]),'T':([0,4,40,],[29,29,29,]),'TD':([0,4,40,],[30,30,30,]),'H':([0,4,40,],[31,31,31,]),'CZ':([0,4,40,],[32,32,32,]),'CNOT':([0,4,40,],[33,33,33,]),'SWP':([0,4,40,],[34,34,34,]),'SSWP':([0,4,40,],[35,35,35,]),'ISWP':([0,4,40,],[36,36,36,]),'SISWP':([0,4,40,],[37,37,37,]),'CP':([0,4,40,],[38,38,38,]),'FSIM':([0,4,40,],[39,39,39,]),'QREG':([5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,43,70,71,72,73,74,75,76,77,],[43,43,43,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,43,88,89,90,91,92,93,94,95,]),'FLOAT':([46,47,48,49,50,51,52,53,80,81,94,95,],[80,81,82,83,84,85,86,87,96,97,98,99,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'root':([0,],[1,]),'program':([0,4,40,],[2,41,78,]),'instruction':([0,4,40,],[3,3,3,]),'qlist':([5,6,7,43,],[42,44,45,79,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> root","S'",1,None,None,None),
  ('root -> program','root',1,'p_root','QCIS_parser.py',59),
  ('program -> instruction NEWLINE program','program',3,'p_program','QCIS_parser.py',63),
  ('program -> instruction','program',1,'p_program','QCIS_parser.py',64),
  ('program -> NEWLINE program','program',2,'p_program','QCIS_parser.py',65),
  ('program -> <empty>','program',0,'p_program','QCIS_parser.py',66),
  ('instruction -> MEASURE qlist','instruction',2,'p_instruction_n_qubit','QCIS_parser.py',71),
  ('instruction -> M qlist','instruction',2,'p_instruction_n_qubit','QCIS_parser.py',72),
  ('instruction -> B qlist','instruction',2,'p_instruction_n_qubit','QCIS_parser.py',73),
  ('qlist -> QREG qlist','qlist',2,'p_qlist','QCIS_parser.py',78),
  ('qlist -> QREG','qlist',1,'p_qlist','QCIS_parser.py',79),
  ('instruction -> XYARB QREG FLOAT FLOAT','instruction',4,'p_instruction_1_qubit_2_param','QCIS_parser.py',89),
  ('instruction -> RXY QREG FLOAT FLOAT','instruction',4,'p_instruction_1_qubit_2_param','QCIS_parser.py',90),
  ('instruction -> RX QREG FLOAT','instruction',3,'p_instruction_1_qubit_1_altitude','QCIS_parser.py',97),
  ('instruction -> RY QREG FLOAT','instruction',3,'p_instruction_1_qubit_1_altitude','QCIS_parser.py',98),
  ('instruction -> XY QREG FLOAT','instruction',3,'p_instruction_1_qubit_1_azimuth','QCIS_parser.py',105),
  ('instruction -> XY2P QREG FLOAT','instruction',3,'p_instruction_1_qubit_1_azimuth','QCIS_parser.py',106),
  ('instruction -> XY2M QREG FLOAT','instruction',3,'p_instruction_1_qubit_1_azimuth','QCIS_parser.py',107),
  ('instruction -> RZ QREG FLOAT','instruction',3,'p_instruction_1_qubit_1_azimuth','QCIS_parser.py',108),
  ('instruction -> X QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',115),
  ('instruction -> X2P QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',116),
  ('instruction -> X2M QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',117),
  ('instruction -> Y QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',118),
  ('instruction -> Y2P QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',119),
  ('instruction -> Y2M QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',120),
  ('instruction -> Z QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',121),
  ('instruction -> Z2P QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',122),
  ('instruction -> Z2M QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',123),
  ('instruction -> Z4P QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',124),
  ('instruction -> Z4M QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',125),
  ('instruction -> S QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',126),
  ('instruction -> SD QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',127),
  ('instruction -> T QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',128),
  ('instruction -> TD QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',129),
  ('instruction -> H QREG','instruction',2,'p_instruction_1_qubit_0_param','QCIS_parser.py',130),
  ('instruction -> CZ QREG QREG','instruction',3,'p_instruction_2_qubit_0_param','QCIS_parser.py',135),
  ('instruction -> CNOT QREG QREG','instruction',3,'p_instruction_2_qubit_0_param','QCIS_parser.py',136),
  ('instruction -> SWP QREG QREG','instruction',3,'p_instruction_2_qubit_0_param','QCIS_parser.py',137),
  ('instruction -> SSWP QREG QREG','instruction',3,'p_instruction_2_qubit_0_param','QCIS_parser.py',138),
  ('instruction -> ISWP QREG QREG','instruction',3,'p_instruction_2_qubit_0_param','QCIS_parser.py',139),
  ('instruction -> SISWP QREG QREG','instruction',3,'p_instruction_2_qubit_0_param','QCIS_parser.py',140),
  ('instruction -> CP QREG QREG FLOAT','instruction',4,'p_instruction_2_qubit_1_param','QCIS_parser.py',147),
  ('instruction -> FSIM QREG QREG FLOAT','instruction',4,'p_instruction_2_qubit_1_param','QCIS_parser.py',148),
]



from tequila import *
from pyqcisim.QCIS_inst import QCISOpCode
import numpy as np

SINGLE_QUBIT_CODE_TENSOR_DICT = {
    QCISOpCode.RX: lambda inst: [
        np.cos(inst.altitude / 2),
        -1j * np.sin(inst.altitude / 2),
        -1j * np.sin(inst.altitude / 2),
        np.cos(inst.altitude / 2),
    ],
    QCISOpCode.RY: lambda inst: [
        np.cos(inst.altitude / 2),
        -np.sin(inst.altitude / 2),
        np.sin(inst.altitude / 2),
        np.cos(inst.altitude / 2),
    ],
    QCISOpCode.RZ: lambda inst: [1, 0, 0, np.exp(1j * inst.azimuth)],
    QCISOpCode.H: lambda inst: TENSOR_BASE_GATE_H,
    QCISOpCode.S: lambda inst: TENSOR_BASE_GATE_S,
    QCISOpCode.SD: lambda inst: [1, 0, 0, -1j],
    QCISOpCode.T: lambda inst: TENSOR_BASE_GATE_T,
    QCISOpCode.TD: lambda inst: [1, 0, 0, np.exp(-1j * np.pi / 4)],
    QCISOpCode.X: lambda inst: TENSOR_BASE_GATE_X,
    QCISOpCode.Z: lambda inst: TENSOR_BASE_GATE_Z,
    QCISOpCode.X2P: lambda inst: [i / (2**0.5) for i in [1, -1j, -1j, 1]],
}

TWO_QUBIT_CODE_TENSOR_DICT = {
    QCISOpCode.CNOT: TENSOR_BASE_GATE_CX,
}


class CircuitExecutorTequila:
    def __init__(self, names):
        self._names = names

        # build <qubit_name, qubit_number> map
        self._qubit_name_number_dict = dict()
        for i, name in enumerate(self._names):
            self._qubit_name_number_dict[name] = i

    def reset(self):
        pass

    def execute(self, instructions, num_shots):
        # prepare qubits
        q = qubit(len(self._names))

        msmt_qubits = []

        # apply gates
        for inst in instructions:
            if inst.op_code.is_single_qubit_op():
                gate_tensor = SINGLE_QUBIT_CODE_TENSOR_DICT[inst.op_code](inst)
                _gate = gate(1)
                _gate.mat2mpo(gate_tensor)
                q.apply_1(self._qubit_name_number_dict[inst.qubit], _gate)

            if inst.op_code.is_two_qubit_op():
                gate_tensor = TWO_QUBIT_CODE_TENSOR_DICT[inst.op_code]
                _gate = gate(2)
                _gate.mat2mpo(gate_tensor)

                q.apply_2(
                    self._qubit_name_number_dict[inst.control_qubit],
                    self._qubit_name_number_dict[inst.target_qubit],
                    _gate,
                )

            if inst.op_code.is_measure_op():
                msmt_qubits.extend(
                    list(
                        map(lambda q: self._qubit_name_number_dict[q], inst.qubits_list)
                    )
                )
        num_msmt_qubits = len(msmt_qubits)
        msmt_qubit_names = list(map(lambda q: self._names[q], msmt_qubits))
        msmts = []
        for i in range(num_shots):
            """meas_rej returns result in a dict format: {'11000': 2, '00100': 3}"""
            bit_str_res = list(q.meas_rej(msmt_qubits, 1).keys())[0]
            res = [int(bit_str_res[i], 2) for i in range(num_msmt_qubits)]
            msmts.append(res)
        return (msmt_qubit_names, msmts)



from pyqcisim.QCIS_parser import QCISParser
from pyqcisim.utils import format_result


class PyQCISim(object):
    """
    A Python-based Quantum Control Instrument Set (QCIS) program simulator.

    Attributes:
        _program: The QCIS program to be simulated
        _parser: A QCIS parser
        _circuit_executor: A quantum circuit simulator executing QCISInst
        _compile_success: The error code for compilation
        _instructions: Generated QCIS instructions by parser
        _executed: The flag for whether instructions have been executed (not used yet)
        _backend: Simulator, 'quantumsim' or 'tequila'
    """

    def __init__(self):
        self._program = ""
        self._parser = None
        self._circuit_executor = None

        self._compile_success = False
        self._instructions = []
        self._names = []
        self._executed = False
        self._backend = "quantumsim"

    def setBackend(self, backend):
        self._backend = backend

        if self._backend == "quantumsim":
            pass
        elif self._backend == "tequila":
            pass
        else:
            raise ValueError("found unsupport backend: {}".format(self._backend))

    def compile(self, prog):
        self._program = prog
        self._parser = QCISParser()  # Parser is also refreshed for a new compile

        success, instructions, names = self._parser.parse(self._program)
        if not success:
            print(self._parser.error_list)
            raise ValueError("QCIS parser failed to compile the given QCIS program.")
        self._compile_success = success
        self._instructions = instructions
        self._names = names

    def simulate(self, mode="one_shot", num_shots=1000):
        if self._backend == "quantumsim":
            return self.simulate_quantumsim(mode, num_shots)
        elif self._backend == "tequila":
            return self.simulate_tequila(mode, num_shots)
        else:
            raise ValueError(
                "Invalid backend: '{}' (supported are: 'quantumsim' and 'tequila').".format(
                    self._backend
                )
            )

    def simulate_quantumsim(self, mode="one_shot", num_shots=1000):
        """Simulate the compiled QCIS program by QuantumSim.

        Args:
          - mode (string): the simulation mode to use:
              - "one_shot":
                The output is a tuple of two elements:
                    - the first element is a list of qubit names `names`
                    - the second element is the collection of all measurement results,
                      with the i-th element in each measurement result being that of
                      the i-th qubit in `names`.
                    Note: the results and qubit names will be sorted.
                 An Example result of a simulation with num_shots=3:
                  ```
                  (['Q3', 'Q4', 'Q5', 'Q6', 'Q7'],
                    [[1, 1, 0, 0, 0],
                     [1, 1, 0, 0, 0],
                     [1, 1, 0, 0, 0]])
                  ```
              - "final_state": the simulation result is a two-level dictionary:
                  {
                    'classical': {'Q1': 1, 'Q2': 0},
                    'quantum': (['Q3', 'Q4'], array([0, 1, 0, 0]))
                  }
          - num_shots (int): the number of iterations performed in `one_shot` mode.
        """
        from pyqcisim.circuit_executor import CircuitExecutor

        if not self._compile_success:
            raise ValueError("Failed to simulate due to compilation error.")

        self._circuit_executor = CircuitExecutor(self._names)

        if mode == "one_shot":
            one_shot_msmts = []
            for i in range(num_shots):
                self._circuit_executor.reset()  # Reset the quantum state simulator

                for inst in self._instructions:  # Execute instructions one by one.
                    self._circuit_executor.execute(inst)

                # Pending operations will not be done!
                # As they do not contribute to the measurement results!
                one_shot_msmts.append(self._circuit_executor.res)

            # final_result = count_final_result(one_shot_msmts)
            final_result = format_result(one_shot_msmts)

        elif mode == "final_state":
            self._circuit_executor.reset()  # Reset the quantum state simulator
            for inst in self._instructions:  # Execute instructions one by one.
                self._circuit_executor.execute(inst)
            final_result = self._circuit_executor.get_quantum_state()

        else:
            raise ValueError(
                "Invalid simulation mode: '{}' (supported are: 'one_shot' and "
                "'final_state').".format(mode)
            )

        return final_result

    def simulate_tequila(self, mode="one_shot", num_shots=1000):
        """Simulate the compiled QCIS program by Tequila."""

        from pyqcisim.circuit_executor_tequila import CircuitExecutorTequila

        if not self._compile_success:
            raise ValueError("Failed to simulate due to compilation error.")
        if not (mode == "one_shot"):
            raise ValueError(
                "Invalid simulation mode: '{}' (supported is: 'one_shot')".format(mode)
            )

        self._circuit_executor = CircuitExecutorTequila(self._names)
        self._circuit_executor.reset()
        ret = self._circuit_executor.execute(self._instructions, num_shots)
        return ret



from argparse import ArgumentParser
from pathlib import Path
from pyqcisim.simulator import PyQCISim
import numpy as np

parser = ArgumentParser(description="QCIS simulator with a QuantumSim backend")

# the input is a positional argument, which is required
parser.add_argument("input", type=str, help="the name of the input QCIS file")

# the output file is an optional argument
parser.add_argument(
    "-m",
    "--mode",
    required=False,
    type=str,
    choices=["one_shot", "final_state"],
    help="the simulation mode used to simulate the given file." " Default to one_shot.",
)

# the output file is an optional argument
parser.add_argument(
    "-n",
    "--num_shots",
    required=False,
    type=int,
    default=1,
    help="the number of iterations performed in the `one_shot` mode.",
)

parser.add_argument(
    "-b",
    "--backend",
    required=False,
    type=str,
    choices=["quantumsim", "tequila"],
    help="the backend to use for simulation. Default to quantumsim.",
)

args = parser.parse_args()

# the input OpenQASM file name
qcis_fn = Path(args.input).resolve()

if qcis_fn.suffix != ".qcis":
    raise ValueError("Error: the input file name should end with the suffix '.qcis'.")


if not qcis_fn.exists():
    raise ValueError("cannot find the given file: {}".format(qcis_fn))
prog = qcis_fn.open("r").read()

pyqcisim = PyQCISim()
pyqcisim.compile(prog)
pyqcisim.setBackend(args.backend)
print(pyqcisim._backend)

if args.mode == "one_shot":
    one_shot_res = pyqcisim.simulate(mode="one_shot", num_shots=args.num_shots)
    np.set_printoptions(precision=3, suppress=True)
    print(one_shot_res)
else:
    final_state = pyqcisim.simulate(mode="final_state")
    np.set_printoptions(precision=3, suppress=True)
    print("qubits measured:", final_state["classical"])
    quantum = final_state["quantum"]
    print("qubits not measured: ", quantum[0])
    print("state held by qubits: ", quantum[1])
    print("absolute: ", np.absolute(quantum[1]))



import numpy as np
from pyqcisim.QCIS_lexer import QCISLexer
import ply.yacc as yacc
from pyqcisim.QCIS_inst import QCISOpCode, QCISInst

OP_CODE_CONVERTER = {
    "MEASURE": QCISOpCode.MEASURE,
    "M": QCISOpCode.M,
    "B": QCISOpCode.B,
    "XY": QCISOpCode.XY,
    "XY2P": QCISOpCode.XY2P,
    "XY2M": QCISOpCode.XY2M,
    "X": QCISOpCode.X,
    "X2P": QCISOpCode.X2P,
    "X2M": QCISOpCode.X2M,
    "Y": QCISOpCode.Y,
    "Y2P": QCISOpCode.Y2P,
    "Y2M": QCISOpCode.Y2M,
    "Z": QCISOpCode.Z,
    "Z2P": QCISOpCode.Z2P,
    "Z2M": QCISOpCode.Z2M,
    "Z4P": QCISOpCode.Z4P,
    "Z4M": QCISOpCode.Z4M,
    "S": QCISOpCode.S,
    "SD": QCISOpCode.SD,
    "T": QCISOpCode.T,
    "TD": QCISOpCode.TD,
    "H": QCISOpCode.H,
    "RX": QCISOpCode.RX,
    "RY": QCISOpCode.RY,
    "RZ": QCISOpCode.RZ,
    "XYARB": QCISOpCode.XYARB,
    "RXY": QCISOpCode.RXY,
    "CZ": QCISOpCode.CZ,
    "CNOT": QCISOpCode.CNOT,
    "SWP": QCISOpCode.SWP,
    "SSWP": QCISOpCode.SSWP,
    "ISWP": QCISOpCode.ISWP,
    "SISWP": QCISOpCode.SISWP,
    "CP": QCISOpCode.CP,
    "FSIM": QCISOpCode.FSIM,
}


class QCISParser(object):
    """QCIS parser"""

    start = "root"

    def __init__(self):
        self.lexer = QCISLexer()
        self.tokens = self.lexer.tokens  # Import all defined tokens in lexer
        self.parser = yacc.yacc(module=self, debug=False)
        self.__instructions = []  # To save all instructions extracted from program
        self.__qubit_names = set()  # A set for all qubit names occurred in program

    # As instructions are generated when specific rules are matched, these two method do not need to do any thing!
    def p_root(self, p):
        """root : program"""
        pass

    def p_program(self, p):
        """program : instruction NEWLINE program
        | instruction
        | NEWLINE program
        |
        """
        pass

    def p_instruction_n_qubit(self, p):
        """instruction : MEASURE qlist
        | M qlist
        | B qlist"""
        self.__qubit_names.update(p[2])
        self.__instructions.append(QCISInst(OP_CODE_CONVERTER[p[1]], qubits_list=p[2]))

    def p_qlist(self, p):
        """qlist : QREG qlist
        | QREG
        """
        if len(p) == 3:  # Matching rule 1
            # Extending qlist.value (a list, p[2]) with QREG.value (a number, p[1])
            p[0] = [p[1]] + p[2]
        else:  # Matching rule 2
            # Create the starting qubits list for all qubits to be measured
            p[0] = [p[1]]

    def p_instruction_1_qubit_2_param(self, p):
        """instruction : XYARB QREG FLOAT FLOAT
        | RXY QREG FLOAT FLOAT"""
        self.__qubit_names.add(p[2])
        self.__instructions.append(
            QCISInst(OP_CODE_CONVERTER[p[1]], qubit=p[2], azimuth=p[3], altitude=p[4])
        )

    def p_instruction_1_qubit_1_altitude(self, p):
        """instruction : RX QREG FLOAT
        | RY QREG FLOAT"""
        self.__qubit_names.add(p[2])
        self.__instructions.append(
            QCISInst(OP_CODE_CONVERTER[p[1]], qubit=p[2], altitude=p[3])
        )

    def p_instruction_1_qubit_1_azimuth(self, p):
        """instruction : XY QREG FLOAT
        | XY2P QREG FLOAT
        | XY2M QREG FLOAT
        | RZ QREG FLOAT"""
        self.__qubit_names.add(p[2])
        self.__instructions.append(
            QCISInst(OP_CODE_CONVERTER[p[1]], qubit=p[2], azimuth=p[3])
        )

    def p_instruction_1_qubit_0_param(self, p):
        """instruction : X QREG
        | X2P QREG
        | X2M QREG
        | Y QREG
        | Y2P QREG
        | Y2M QREG
        | Z QREG
        | Z2P QREG
        | Z2M QREG
        | Z4P QREG
        | Z4M QREG
        | S QREG
        | SD QREG
        | T QREG
        | TD QREG
        | H QREG"""
        self.__qubit_names.add(p[2])
        self.__instructions.append(QCISInst(OP_CODE_CONVERTER[p[1]], qubit=p[2]))

    def p_instruction_2_qubit_0_param(self, p):
        """instruction : CZ QREG QREG
        | CNOT QREG QREG
        | SWP QREG QREG
        | SSWP QREG QREG
        | ISWP QREG QREG
        | SISWP QREG QREG"""
        self.__qubit_names.update({p[2], p[3]})
        self.__instructions.append(
            QCISInst(OP_CODE_CONVERTER[p[1]], control_qubit=p[2], target_qubit=p[3])
        )

    def p_instruction_2_qubit_1_param(self, p):
        """instruction : CP QREG QREG FLOAT
        | FSIM QREG QREG FLOAT"""
        self.__qubit_names.update({p[2], p[3]})
        self.__instructions.append(
            QCISInst(
                OP_CODE_CONVERTER[p[1]],
                control_qubit=p[2],
                target_qubit=p[3],
                azimuth=p[4],
            )
        )

    def p_error(self, p):
        """Error handling method skipping error-occurred line."""
        if not p:
            error_msg = "Syntax error at the end of program."
            self.error_list.append(error_msg)
            return

        col = self.lexer.find_column(p)
        lines = self.lexer.data.split("\n")
        error_msg = "Syntax error: Found unmatched {0} at ({1}, {2}).\nSkip line {1}: {3}".format(
            p.type, self.lexer.lineno, col, lines[p.lineno - 1]
        )
        self.error_list.append(error_msg)

        # Read ahead looking for a new line
        while True:
            tok = self.lexer.token()  # Get the next token
            if not tok or tok.type == "NEWLINE":
                break
        self.parser.restart()

    def parse(self, data):
        self.__instructions = []  # Clear instructions list
        self.__qubit_names = set()  # Clear qubit names set
        self.error_list = []
        self.lexer = QCISLexer()  # Reset initial lineno of lexer
        self.parser.parse(data, lexer=self.lexer)

        success = True
        if len(self.error_list) > 0:
            error_msgs = "\n".join(self.error_list)
            print("Found errors during parsing QCIS program: {}".format(error_msgs))
            success = False

        qubit_names_list = list(self.__qubit_names)  # Turn the set to a list
        qubit_names_list.sort()  # Sort by alphabetic
        return success, self.__instructions, qubit_names_list



import numpy as np
import pytest

REPEAT = 2000  # Repeat times for sampling the measurement result in tests


def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def get_bin(x, n, Unsigned=True):
    """
    Return the 2's complement of the integer number $x$
    for a given bitwidth $n$.
    """
    if not is_number(x):
        raise ValueError("get_bin: parameter {} is not a number.".format(x))

    if Unsigned is True:
        return "{0:{fill}{width}b}".format(int(x), fill="0", width=n)
    else:
        return "{0:{fill}{width}b}".format(
            (int(x) + 2**n) % 2**n, fill="0", width=n
        )


def format_result(final_results):
    """
    Format the one shot simulation results of multiple iterations.

    Input: [{'Q3': 1, 'Q4': 1, 'Q5': 0, 'Q6': 0, 'Q7': 0}, {'Q3': 1, 'Q4': 1, 'Q5': 0, 'Q6': 0, 'Q7': 0}, {'Q3': 1, 'Q4': 1, 'Q5': 0, 'Q6': 0, 'Q7': 0}]

    The output is a tuple of two elements:
    - the first element is a list of qubit names `names`
    - the second element is the collection of all measurement results, with the i-th
      element in each measurement result being that of the i-th qubit in `names`.
    Note: the qubit names will be sorted. Example:
    (['Q3', 'Q4', 'Q5', 'Q6', 'Q7'], [[1, 1, 0, 0, 0], [1, 1, 0, 0, 0], [1, 1, 0, 0, 0]])
    """
    if len(final_results) == 0 or len(final_results[0]) == 0:
        return None

    first_shot_result = final_results[0]
    num_qubits = len(first_shot_result)
    names = list(first_shot_result.keys())
    names.sort(key=lambda x: int(x[1:]))

    assert all(isinstance(one_shot, dict) for one_shot in final_results)
    assert all([len(one_shot) == num_qubits for one_shot in final_results])

    result_matrix = []
    for one_shot in final_results:
        res = [one_shot[name] for name in names]
        result_matrix.append(res)

    return (names, result_matrix)


def count_final_result(final_results):
    if len(final_results) == 0:
        return None

    num_qubits = len(final_results[0])
    if num_qubits == 0:
        return None

    assert all([len(one_result) == num_qubits for one_result in final_results])

    name_list = []
    for name in final_results[0]:
        name_list.append(name)

    result_vector = [0] * (2**num_qubits)

    for one_round_msmt in final_results:
        bin_addr = 0
        shift = 0
        for qubit_name in name_list:
            bin_addr += one_round_msmt[qubit_name] << shift
            shift += 1
        result_vector[bin_addr] += 1
    res_dict = {}
    for i in range(len(result_vector)):
        res_dict[get_bin(i, num_qubits)] = result_vector[i]

    return (name_list, res_dict)


def state_vec_equal(state_vec1, state_vec2):
    """Compare two vectors with each presenting a quantum state vector are equal."""
    # Find out the norm of the inner product of two quantum state vector
    # `vdot` calculates the inner product of complex vectors
    # `abs` calculates the norm of a complex number
    norm = abs(np.vdot(state_vec1, state_vec2))
    # Norm is 1 only when these two quantum state vector are physically identical
    return norm == pytest.approx(1)


def quantum_state_vec_equal(quantum_state1, quantum_state2):
    """Compare two quantum states are equal. They are equal iff:
    - the list of qubit names are identical
    - the corresponding state vectors are equal.
    """
    # compare the list of qubit names
    same_qname = quantum_state1[0] == quantum_state2[0]

    # compare the state vector
    same_qvec = state_vec_equal(quantum_state1[1], quantum_state2[1])
    return same_qname and same_qvec


def seperate_state_cmp(state1, state2):
    """Compare two seperated output quantum state in tests."""
    same_classical = state1["classical"] == state2["classical"]
    same_quantum = quantum_state_vec_equal(state1["quantum"], state2["quantum"])
    return same_classical and same_quantum


def stats_cmp(stats_res, target, precision=2e-2):
    """Compare a statistic measurement result with a target distribution."""
    keys = [key for key in stats_res.keys()]
    assert all(key in keys for key in target.keys())

    assert sum(target[key] for key in target.keys()) == pytest.approx(1, abs=0.001)

    total_count = sum([stats_res[key] for key in keys])
    for key in target.keys():
        if not stats_res[key] / total_count == pytest.approx(
            target[key], abs=precision
        ):
            return False
    return True


def rotate_x_mat(theta):
    """Return unitary matrix rotating state vector along x-axis clockwisely."""
    return np.array(
        [
            [np.cos(theta / 2), -1.0j * np.sin(theta / 2)],
            [-1.0j * np.sin(theta / 2), np.cos(theta / 2)],
        ]
    )


def rotate_y_mat(theta):
    """Return unitary matrix rotating state vector along y-axis clockwisely."""
    return np.array(
        [
            [np.cos(theta / 2), -np.sin(theta / 2)],
            [np.sin(theta / 2), np.cos(theta / 2)],
        ]
    )


def rotate_z_mat(phi):
    """Return unitary matrix rotating state vector along z-axis clockwisely."""
    return np.array(
        [
            [1, 0],
            [0, np.exp(1.0j * phi)],
        ]
    )


ZERO_STATE = [1, 0]
ONE_STATE = [0, 1]
P_STATE = np.multiply((1 / np.sqrt(2)), [1, 1])  # State along x-axis
M_STATE = np.multiply((1 / np.sqrt(2)), [1, -1])  # State along -x
L_STATE = np.multiply((1 / np.sqrt(2)), [1, 1.0j])  # State along y-axis
R_STATE = np.multiply((1 / np.sqrt(2)), [1, -1.0j])  # State along -y

BELL_STATE = np.multiply((1 / np.sqrt(2)), [1, 0, 0, 1])



import ply.lex as lex


class QCISLexer(object):
    """QCIS lexer"""

    def __init__(self):
        """Create a ply lexer."""
        self.lexer = lex.lex(module=self, debug=False, errorlog=lex.NullLogger())
        self.lineno = 1

    def input(self, data):
        """Set the input text data."""
        self.data = data.upper()
        self.lexer.input(data.upper())

    def token(self):
        """Return the next token."""
        ret = self.lexer.token()
        return ret

    reserved = {
        # ---Basic Operations---
        "CZ": "CZ",
        "MEASURE": "MEASURE",  # Deprecated measurement, use `M`
        "M": "M",  # Recommended measurement
        "RZ": "RZ",
        "XYARB": "XYARB",  # Deprecated arbitrary xy-plane axis rotation, use `RXY`
        "XY": "XY",
        "XY2P": "XY2P",
        "XY2M": "XY2M",
        "X": "X",
        "X2P": "X2P",
        "X2M": "X2M",
        "Y": "Y",
        "Y2P": "Y2P",
        "Y2M": "Y2M",
        "Z": "Z",
        "Z2P": "Z2P",  # Deprecated z-axis rotation, use `S`
        "Z2M": "Z2M",  # Deprecated z-axis rotation, use `SD`
        "Z4P": "Z4P",  # Deprecated z-axis rotation, use `T`
        "Z4M": "Z4M",  # Deprecated z-axis rotation , use `TD`
        "S": "S",
        "SD": "SD",
        "T": "T",
        "TD": "TD",
        "B": "B",  # Time barrier for two qubits (useless in simulator)
        # ---Extended Operations---
        "H": "H",
        "RX": "RX",
        "RY": "RY",
        "RXY": "RXY",
        "CNOT": "CNOT",
        "SWP": "SWP",
        "SSWP": "SSWP",
        "ISWP": "ISWP",
        "SISWP": "SISWP",
        "CP": "CP",
        "FSIM": "FSIM",
        # ---Advanced Operations---
        #! Note: Not supported now!
        # "SET": "SET",
        # "CMT": "CMT",
        # "I": "I",
        # "X12": "X12",
        # "X23": "X23",
        # "AXY": "AXY",
        # "DTN": "DTN",
        # "PLS": "PLS",
        # "PLSXY": "PLSXY",
        # "SWD": "SWD",
        # "SWA": "SWA",
        # "MOV": "MOV",
    }

    tokens = ["FLOAT", "QREG", "NEWLINE"] + list(reserved.values())

    def t_FLOAT(self, t):
        r"[-+]?[0-9]+(\.([0-9]+)?([eE][-+]?[0-9]+)?|[eE][-+]?[0-9]+)?"
        # r"[-]?\d+(\.\d+)?"
        t.value = float(t.value)

        return t

    def t_QREG(self, t):
        r"[a-zA-Z_][\.a-zA-Z_0-9]*"
        # If it is an identifier but not a reserved word, then it must be QREG
        t.type = self.reserved.get(t.value, "QREG")
        return t

    def t_NEWLINE(self, t):
        r"\n"
        self.lineno += len(t.value)
        t.lexer.lineno = self.lineno
        return t

    t_ignore = " \t"
    # Note that `.` can not match \n
    t_ignore_COMMENT = r"\#.*"

    def find_column(self, t):
        """Compute the column of the token t."""
        if t is None:
            return 0
        last_line_end = self.data.rfind("\n", 0, t.lexpos)
        column = t.lexpos - last_line_end
        return column

    def t_error(self, t):
        # When a token has no matching rule, `t.value` will contain the rest of input.
        # Thus we usually only report the string before the first space, \n or \t.
        raise ValueError(
            "Give string ({}) at (line {}, col {}) cannot match any QCIS token rule".format(
                t.value.split(" ")[0].split("\n")[0].split("\t")[0],
                t.lexer.lineno,
                self.find_column(t),
            )
        )

    def get_all_token(self):
        return [tok for tok in self.lexer]

    def get_all_token_info(self):
        return [(tok.type, tok.value, tok.lineno, tok.lexpos) for tok in self.lexer]



import numpy as np
from quantumsim.sparsedm import *
from quantumsim.circuit import *
from quantumsim.ptm import *
from pyqcisim.QCIS_inst import QCISOpCode
import random

# TODO This part has became complicated. Refactor is needed.
SINGLE_QUBIT_INST_DECODE_HELPER = {
    QCISOpCode.RZ: [
        (rotate_z_ptm, lambda inst: (inst.azimuth,)),
    ],
    QCISOpCode.XYARB: [
        (rotate_z_ptm, lambda inst: (-inst.azimuth,)),
        (rotate_x_ptm, lambda inst: (inst.altitude,)),
        (rotate_z_ptm, lambda inst: (inst.azimuth,)),
    ],
    QCISOpCode.XY: [
        (rotate_z_ptm, lambda inst: (-inst.azimuth,)),
        (rotate_x_ptm, lambda inst: (np.pi,)),
        (rotate_z_ptm, lambda inst: (inst.azimuth,)),
    ],
    QCISOpCode.XY2P: [
        (rotate_z_ptm, lambda inst: (-inst.azimuth,)),
        (rotate_x_ptm, lambda inst: (np.pi / 2,)),
        (rotate_z_ptm, lambda inst: (inst.azimuth,)),
    ],
    QCISOpCode.XY2M: [
        (rotate_z_ptm, lambda inst: (-inst.azimuth,)),
        (rotate_x_ptm, lambda inst: (-np.pi / 2,)),
        (rotate_z_ptm, lambda inst: (inst.azimuth,)),
    ],
    QCISOpCode.X: [
        (rotate_x_ptm, lambda inst: (np.pi,)),
    ],
    QCISOpCode.X2P: [
        (rotate_x_ptm, lambda inst: (np.pi / 2,)),
    ],
    QCISOpCode.X2M: [
        (rotate_x_ptm, lambda inst: (-np.pi / 2,)),
    ],
    QCISOpCode.Y: [
        (rotate_y_ptm, lambda inst: (np.pi,)),
    ],
    QCISOpCode.Y2P: [
        (rotate_y_ptm, lambda inst: (np.pi / 2,)),
    ],
    QCISOpCode.Y2M: [
        (rotate_y_ptm, lambda inst: (-np.pi / 2,)),
    ],
    QCISOpCode.Z: [
        (rotate_z_ptm, lambda inst: (np.pi,)),
    ],
    QCISOpCode.Z2P: [
        (rotate_z_ptm, lambda inst: (np.pi / 2,)),
    ],
    QCISOpCode.Z2M: [
        (rotate_z_ptm, lambda inst: (-np.pi / 2,)),
    ],
    QCISOpCode.Z4P: [
        (rotate_z_ptm, lambda inst: (np.pi / 4,)),
    ],
    QCISOpCode.Z4M: [
        (rotate_z_ptm, lambda inst: (-np.pi / 4,)),
    ],
    QCISOpCode.S: [
        (rotate_z_ptm, lambda inst: (np.pi / 2,)),
    ],
    QCISOpCode.SD: [
        (rotate_z_ptm, lambda inst: (-np.pi / 2,)),
    ],
    QCISOpCode.T: [
        (rotate_z_ptm, lambda inst: (np.pi / 4,)),
    ],
    QCISOpCode.TD: [
        (rotate_z_ptm, lambda inst: (-np.pi / 4,)),
    ],
    QCISOpCode.H: [
        (hadamard_ptm, lambda inst: ()),
    ],
    QCISOpCode.RX: [
        (rotate_x_ptm, lambda inst: (inst.altitude,)),
    ],
    QCISOpCode.RY: [
        (rotate_y_ptm, lambda inst: (inst.altitude,)),
    ],
    QCISOpCode.RXY: [
        (rotate_z_ptm, lambda inst: (-inst.azimuth,)),
        (rotate_x_ptm, lambda inst: (inst.altitude,)),
        (rotate_z_ptm, lambda inst: (inst.azimuth,)),
    ],
}

TWO_QUBIT_DECODE_HELPER = {
    QCISOpCode.CZ: ptm.double_kraus_to_ptm(np.diag([1, 1, 1, -1])),
    QCISOpCode.CNOT: ptm.double_kraus_to_ptm(
        np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])
    ),
    QCISOpCode.CP: [
        ptm.double_kraus_to_ptm,
        lambda inst: np.diag([1, 1, 1, np.exp(1j * inst.azimuth)]),
    ],
    QCISOpCode.SWP: ptm.double_kraus_to_ptm(
        np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])
    ),
    QCISOpCode.SSWP: ptm.double_kraus_to_ptm(
        np.array(
            [
                [1, 0, 0, 0],
                [0, (1 + 1.0j) / 2, (1 - 1.0j) / 2, 0],
                [0, (1 - 1.0j) / 2, (1 + 1.0j) / 2, 0],
                [0, 0, 0, 1],
            ]
        )
    ),
    QCISOpCode.ISWP: ptm.double_kraus_to_ptm(
        np.array([[1, 0, 0, 0], [0, 0, 1.0j, 0], [0, 1.0j, 0, 0], [0, 0, 0, 1]])
    ),
    QCISOpCode.SISWP: ptm.double_kraus_to_ptm(
        np.array(
            [
                [1, 0, 0, 0],
                [0, 1 / np.sqrt(2), 1.0j / np.sqrt(2), 0],
                [0, 1.0j / np.sqrt(2), 1 / np.sqrt(2), 0],
                [0, 0, 0, 1],
            ]
        )
    ),
}


class CircuitExecutor(object):
    def __init__(self, names):
        self._names = names
        self._sdm = SparseDM(names)
        self.res = []

    def reset(self):
        """
        Reset the qubit state simulator by creating a new instance of the density matrix.

        In this way, the following simulation is completely independent of previous ones.
        """
        self._sdm = SparseDM(self._names)
        self.res = []
        rng = np.random.RandomState(seed=42)

    def execute(self, inst):
        if inst.op_code.is_single_qubit_op():
            for ptm, param in SINGLE_QUBIT_INST_DECODE_HELPER[inst.op_code]:
                self._sdm.apply_ptm(inst.qubit, ptm(*param(inst)))
                self._sdm.ensure_dense(inst.qubit)
                self._sdm.combine_and_apply_single_ptm(inst.qubit)
            return

        if inst.op_code.is_two_qubit_op():
            if inst.op_code.is_two_qubit_param_op():
                ptm_generator, param = TWO_QUBIT_DECODE_HELPER[inst.op_code]
                what_get = param(inst)
                ptm = ptm_generator(what_get)
            else:
                ptm = TWO_QUBIT_DECODE_HELPER[inst.op_code]

            self._sdm.ensure_dense(inst.control_qubit)
            self._sdm.ensure_dense(inst.target_qubit)
            # These additional real applying is needed, as `apply_two_ptm` doesn't do these
            self._sdm.combine_and_apply_single_ptm(inst.control_qubit)
            self._sdm.combine_and_apply_single_ptm(inst.target_qubit)

            #! The first parameter postion is for target qubit, according to `test_cnot`
            self._sdm.apply_two_ptm(inst.target_qubit, inst.control_qubit, ptm)
            return

        if inst.op_code.is_measure_op():
            for qubit in inst.qubits_list:
                if qubit in self._sdm.classical:
                    continue

                p0, p1 = self._sdm.peak_measurement(qubit)
                random.seed()  # seed from current time for full randomness.
                r = random.random()
                if r < p0 / (p0 + p1):
                    project = 0
                else:
                    project = 1

                self._sdm.project_measurement(qubit, project)
                self._sdm.renormalize()
                # self._sdm.set_bit(self.real_output_bit, project)

            self.res = self._sdm.classical
            return

        if inst.op_code == QCISOpCode.B:
            return

        raise ValueError("Given instruction {} can not be decoded".format(inst))

    def get_quantum_state(self, seperate=True):
        """Take the quantum state before measurements."""
        # Apply all rest cached single qubit gates
        self._sdm.apply_all_pending()

        if seperate:
            return {
                "classical": self._sdm.classical,
                "quantum": self._dm_to_vec(self._sdm.full_dm, self._sdm.idx_in_full_dm),
            }
        else:
            for qubit in self._sdm.names:
                self._sdm.ensure_dense(qubit)
            return self._dm_to_vec(self._sdm.full_dm, self._sdm.idx_in_full_dm)

    def _dm_to_vec(self, full_dm, idx_in_full_dm):
        """Method turning a pure state density matrix into state vector."""
        dm_array = full_dm.to_array()
        assert (
            2 ** (len(idx_in_full_dm.items())) == dm_array.shape[0] == dm_array.shape[1]
        )

        # The case where all qubits are non-quantum and dm is empty
        if idx_in_full_dm == {}:
            return [], 1  # Return empty qubits list and a trivial number

        num_col = dm_array.shape[1]  # Number of columns
        # The target state vector got from normalizing the column with the largest norm.
        # Mathematically, a colunm with non-zero norm is enough, but in real computer *some almost zero number is not zero*.
        # Thus, we choose the-largest-norm column to get away from that situation.
        state_vec = None
        cur_max_norm = 0
        for i in range(0, num_col):
            col_vec = dm_array[:, i]
            # This norm is just the norm of the current col's common (one-of-amplitude)* factor
            norm = np.linalg.norm(col_vec)
            if norm > cur_max_norm:
                state_vec = col_vec / norm
                cur_max_norm = norm

        # Get the corresponding qubit names list for the state vector
        names_indices = list(zip(idx_in_full_dm.keys(), idx_in_full_dm.values()))
        names_indices.sort(key=lambda pair: pair[1])

        names_seq = [pair[0] for pair in names_indices]
        return names_seq, state_vec



from enum import Enum, auto


class QCISOpCode(Enum):
    # The first single-qubit operation
    RZ = auto()
    XYARB = auto()
    XY = auto()
    XY2P = auto()
    XY2M = auto()
    X = auto()
    X2P = auto()
    X2M = auto()
    Y = auto()
    Y2P = auto()
    Y2M = auto()
    Z = auto()
    Z2P = auto()
    Z2M = auto()
    Z4P = auto()
    Z4M = auto()
    S = auto()
    SD = auto()
    T = auto()
    TD = auto()
    H = auto()
    RX = auto()
    RY = auto()
    RXY = auto()
    # The last single-qubit operation

    # The first two-qubit operation
    CZ = auto()
    CNOT = auto()
    SWP = auto()
    SSWP = auto()
    ISWP = auto()
    SISWP = auto()
    # The last two-qubit operation

    # The first two-qubit operation with a parameter
    CP = auto()
    FSIM = auto()
    # The last two-qubit operation with a parameter

    # The first measurement operation
    MEASURE = auto()
    M = auto()
    # The last measurement operation

    B = auto()

    def is_single_qubit_op(self):
        return self.RZ.value <= self.value <= self.RXY.value

    def is_two_qubit_op(self):
        return self.CZ.value <= self.value <= self.FSIM.value

    def is_two_qubit_param_op(self):
        return self.CP.value <= self.value <= self.FSIM.value

    def is_measure_op(self):
        return self.MEASURE.value <= self.value <= self.M.value


class QCISInst(object):
    def __init__(self, op_code, **kwargs):
        """
        Data structure for representing QCIS instructions.

        Attributes:
            op_code: The operation code of the QCIS instruction.
            azimuth: The angle between the axis to rotate along and z-axis.
            altitude: The angle of rotation along a given axis.

        Single-qubit operation only attributes:
            qubit: The name string of target qubit.

        Two-qubit operation only attributes:
            control_qubit: The name string of control qubit.
            target_qubit: The name string of target qubit.

        Measurement operation only attributes:
            qubits_list: The names of all qubits to be measured.
        """
        self.op_code = op_code

        # TODO This part is awkward. Refactor is needed!
        if op_code.is_two_qubit_op():
            if self.op_code == QCISOpCode.CP or self.op_code == QCISOpCode.FSIM:
                self.azimuth = kwargs["azimuth"]
            self.control_qubit = kwargs["control_qubit"]
            self.target_qubit = kwargs["target_qubit"]
            return

        if op_code.is_single_qubit_op():
            self.qubit = kwargs["qubit"]

            if self.op_code == QCISOpCode.XYARB or self.op_code == QCISOpCode.RXY:
                self.azimuth = kwargs["azimuth"]
                self.altitude = kwargs["altitude"]
                return

            if (
                self.op_code == QCISOpCode.XY
                or self.op_code == QCISOpCode.XY2P
                or self.op_code == QCISOpCode.XY2M
                or self.op_code == QCISOpCode.RZ
            ):
                self.azimuth = kwargs["azimuth"]
                return

            if self.op_code == QCISOpCode.RX or self.op_code == QCISOpCode.RY:
                self.altitude = kwargs["altitude"]
                return

            return

        if op_code.is_measure_op():
            # Should be a list even measuring only one qubit
            self.qubits_list = kwargs["qubits_list"]
            self.qubits_list.sort()
            return

        if op_code == QCISOpCode.B:
            self.qubits_list = kwargs["qubits_list"]
            self.qubits_list.sort()
            return

        raise ValueError("Found unrecognized opcode: ", op_code)

    def __str__(self):
        # TODO Update this method after refactoring this class
        if self.op_code.is_two_qubit_op():
            return "Two-qubit op: {}, control: {}, target: {}".format(
                self.op_code, self.control_qubit, self.target_qubit
            )

        if self.op_code.is_single_qubit_op():
            params_str = ""
            if self.op_code == QCISOpCode.XYARB:
                params_str = ", azimuth: {}, altitude: {}".format(
                    self.azimuth, self.altitude
                )
            return "Single-qubit op: {}, qubit: {}{}".format(
                self.op_code, self.qubit, params_str
            )

        if self.op_code.is_measure_op():
            qubits_list_str = " ".join([qubit for qubit in self.qubits_list])
            return "Measure op: {}, qubits list: {}".format(
                self.op_code, qubits_list_str
            )

        raise ValueError("Unrecognized instruction.")

    def __eq__(self, other):
        # Two QCISInst instances with same values of attributes will be identical
        return self.__dict__ == other.__dict__
